@@ -4,7 +4,7 @@ import numpy as np
 import os
 from numba import njit, prange
 
-from .maxcut_tfim_util import fix_cdf, get_cut, init_thresholds, maxcut_hamming_cdf, opencl_context, probability_by_hamming_weight
+from .maxcut_tfim_util import fix_cdf, get_cut, init_thresholds, maxcut_hamming_cdf, opencl_context
 
 IS_OPENCL_AVAILABLE = True
 try:
@@ -22,7 +22,7 @@ def update_repulsion_choice(G_m, max_weight, weights, n, used, node):
     for nbr in range(n):
         if used[nbr]:
             continue
-        weights[nbr] *= max(2e-7, 1 - G_m[node, nbr] / max_weight)
+        weights[nbr] *= max(1.1920928955078125e-7, 1 - G_m[node, nbr] / max_weight)
 
 
 # Written by Elara (OpenAI custom GPT) and improved by Dan Strano
@@ -33,7 +33,7 @@ def local_repulsion_choice(G_m, max_weight, weights, n, m, shot):
     - High-degree nodes are already less likely
     - After choosing a node, its neighbors' probabilities are further reduced
     adjacency_data, adjacency_rows: CSR-format sparse adjacency data
-    weights: float64 array of shape (n,)
+    weights: float32 array of shape (n,)
     """
 
     weights = weights.copy()
@@ -84,13 +84,12 @@ def compute_energy(sample, G_m, n_qubits):
 
 
 @njit(parallel=True)
-def sample_for_solution(G_m, shots, thresholds, J_eff):
+def sample_for_solution(G_m, shots, thresholds, weights, dtype):
     n = len(G_m)
     max_weight = G_m.max()
-    weights = (1.0 / (1.0 + (2e-52) - J_eff)).astype(np.float64)
 
     solutions = np.empty((shots, n), dtype=np.bool_)
-    energies = np.empty(shots, dtype=np.float32)
+    energies = np.empty(shots, dtype=dtype)
 
     for s in prange(shots):
         # First dimension: Hamming weight
@@ -118,10 +117,10 @@ def sample_for_solution(G_m, shots, thresholds, J_eff):
 
 
 @njit(parallel=True)
-def init_J_and_z(G_m):
+def init_J_and_z(G_m, dtype):
     n_qubits = len(G_m)
     degrees = np.empty(n_qubits, dtype=np.uint32)
-    J_eff = np.empty(n_qubits, dtype=np.float32)
+    J_eff = np.empty(n_qubits, dtype=dtype)
     J_max = -float("inf")
     for n in prange(n_qubits):
         degree = sum(G_m[n] != 0.0)
@@ -139,6 +138,7 @@ def run_sampling_opencl(G_m_np, thresholds_np, shots, n, is_g_buf_reused):
     ctx = opencl_context.ctx
     queue = opencl_context.queue
     kernel = opencl_context.sample_for_solution_best_bitset_kernel
+    dtype = opencl_context.dtype
 
     max_local_size = 64  # tune
     max_global_size = ((opencl_context.MAX_GPU_PROC_ELEM + max_local_size - 1) // max_local_size) * max_local_size  # corresponds to MAX_PROC_ELEM macro in OpenCL kernel program
@@ -163,11 +163,11 @@ def run_sampling_opencl(G_m_np, thresholds_np, shots, n, is_g_buf_reused):
 
     # Best energies and solutions (per group)
     # We'll reuse solutions_buf to hold them after reduction (kernel copies winners into group slices)
-    best_energies_np = np.empty(num_groups, dtype=np.float32)
+    best_energies_np = np.empty(num_groups, dtype=dtype)
     best_energies_buf = cl.Buffer(ctx, mf.WRITE_ONLY, best_energies_np.nbytes)
 
     # Local memory buffers
-    local_energy_buf = cl.LocalMemory(np.dtype(np.float32).itemsize * local_size)
+    local_energy_buf = cl.LocalMemory(np.dtype(dtype).itemsize * local_size)
     local_index_buf = cl.LocalMemory(np.dtype(np.int32).itemsize * local_size)
 
     # Set kernel args
@@ -176,7 +176,7 @@ def run_sampling_opencl(G_m_np, thresholds_np, shots, n, is_g_buf_reused):
         thresholds_buf,
         np.int32(n),
         np.int32(shots),
-        np.float64(G_m_np.max()),
+        dtype(G_m_np.max()),
         rng_buf,
         solutions_buf,
         best_energies_buf,
@@ -212,13 +212,16 @@ def run_sampling_opencl(G_m_np, thresholds_np, shots, n, is_g_buf_reused):
 
 
 @njit
-def cpu_footer(shots, quality, n_qubits, G_m, nodes):
-    J_eff, degrees = init_J_and_z(G_m)
-    hamming_prob = init_thresholds(n_qubits)
+def cpu_footer(shots, quality, n_qubits, G_m, nodes, dtype):
+    J_eff, degrees = init_J_and_z(G_m, dtype)
+    hamming_prob = init_thresholds(n_qubits, dtype)
 
-    maxcut_hamming_cdf(n_qubits, J_eff, degrees, quality, hamming_prob)
+    maxcut_hamming_cdf(n_qubits, J_eff, degrees, quality, hamming_prob, dtype)
 
-    best_solution, best_value = sample_for_solution(G_m, shots, hamming_prob, J_eff)
+    degrees = None
+    J_eff = 1.0 / (1.0 + (2 ** -52) - J_eff)
+
+    best_solution, best_value = sample_for_solution(G_m, shots, hamming_prob, J_eff, dtype)
 
     bit_string, l, r = get_cut(best_solution, nodes)
 
@@ -226,10 +229,10 @@ def cpu_footer(shots, quality, n_qubits, G_m, nodes):
 
 
 @njit
-def gpu_footer(shots, n_qubits, G_m, J_eff, hamming_prob, nodes):
+def gpu_footer(shots, n_qubits, G_m, weights, hamming_prob, nodes, dtype):
     fix_cdf(hamming_prob)
 
-    best_solution, best_value = sample_for_solution(G_m, shots, hamming_prob, J_eff)
+    best_solution, best_value = sample_for_solution(G_m, shots, hamming_prob, weights, dtype)
 
     bit_string, l, r = get_cut(best_solution, nodes)
 
@@ -240,16 +243,18 @@ def maxcut_tfim(
     G,
     quality=None,
     shots=None,
-    is_alt_gpu_sampling = False,
-    is_g_buf_reused = False
+    is_alt_gpu_sampling=False,
+    is_g_buf_reused=False,
+    is_base_maxcut_gpu=True
 ):
+    dtype = opencl_context.dtype
     nodes = None
     n_qubits = 0
     G_m = None
     if isinstance(G, nx.Graph):
         nodes = list(G.nodes())
         n_qubits = len(nodes)
-        G_m = nx.to_numpy_array(G, weight='weight', nonedge=0.0)
+        G_m = nx.to_numpy_array(G, weight='weight', nonedge=0.0, dtype=dtype)
     else:
         n_qubits = len(G)
         nodes = list(range(n_qubits))
@@ -276,19 +281,19 @@ def maxcut_tfim(
         # Number of measurement shots
         shots = n_qubits << quality
 
-    n_steps = 2 << quality
+    n_steps = n_qubits << quality
     grid_size = n_steps * n_qubits
 
-    if not (IS_OPENCL_AVAILABLE and grid_size >= 128):
+    if (not is_base_maxcut_gpu) or not (IS_OPENCL_AVAILABLE and grid_size >= 128):
         return cpu_footer(shots, quality, n_qubits, G_m, nodes)
 
-    J_eff, degrees = init_J_and_z(G_m)
+    J_eff, degrees = init_J_and_z(G_m, dtype)
 
     delta_t = 1.0 / n_steps
     tot_t = 2.0 * n_steps * delta_t
     h_mult = 2.0 / tot_t
 
-    args = np.empty(3, dtype=np.float32)
+    args = np.empty(3, dtype=dtype)
     args[0] = delta_t
     args[1] = tot_t
     args[2] = h_mult
@@ -298,7 +303,7 @@ def maxcut_tfim(
     args_buf = cl.Buffer(opencl_context.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=args)
     J_buf = cl.Buffer(opencl_context.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=J_eff)
     deg_buf = cl.Buffer(opencl_context.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=degrees)
-    theta_buf = cl.Buffer(opencl_context.ctx, mf.READ_WRITE, size=(n_qubits * np.float32().nbytes))
+    theta_buf = cl.Buffer(opencl_context.ctx, mf.READ_WRITE, size=(n_qubits * dtype().nbytes))
 
     # Warp size is 32:
     group_size = min(n_qubits, 64)
@@ -309,7 +314,7 @@ def maxcut_tfim(
         args_buf, np.int32(n_qubits), J_buf, deg_buf, theta_buf
     )
 
-    hamming_prob = init_thresholds(n_qubits)
+    hamming_prob = init_thresholds(n_qubits, dtype)
 
     # Warp size is 32:
     group_size = n_qubits - 1
@@ -341,10 +346,13 @@ def maxcut_tfim(
     theta_buf = None
 
     if not is_alt_gpu_sampling:
-        return gpu_footer(shots, n_qubits, G_m, J_eff, hamming_prob, nodes)
+        degrees = None
+        J_eff = 1.0 / (1.0 + (2 ** -52) - J_eff)
+
+        return gpu_footer(shots, n_qubits, G_m, J_eff, hamming_prob, nodes, dtype)
 
     fix_cdf(hamming_prob)
-    best_solution, best_value = run_sampling_opencl(G_m, hamming_prob, shots, n_qubits, is_g_buf_reused)
+    best_solution, best_value = run_sampling_opencl(G_m, hamming_prob, shots, n_qubits, is_g_buf_reused, dtype)
     bit_string, l, r = get_cut(best_solution, nodes)
 
     return bit_string, best_value, (l, r)
