@@ -22,7 +22,7 @@ def update_repulsion_choice(G_func, G_func_args_tuple, nodes, max_weight, weight
     for nbr in range(n):
         if used[nbr]:
             continue
-        weights[nbr] *= max(2e-7, 1 - G_func((nodes[node], nodes[nbr]), G_func_args_tuple) / max_weight)
+        weights[nbr] *= max(1.1920928955078125e-7, 1 - G_func((nodes[node], nodes[nbr]), G_func_args_tuple) / max_weight)
 
 
 # Written by Elara (OpenAI custom GPT) and improved by Dan Strano
@@ -87,11 +87,9 @@ def compute_energy(sample, G_func, G_func_args_tuple, nodes, n_qubits):
 
 
 @njit(parallel=True)
-def sample_for_solution(G_func, G_func_args_tuple, nodes, max_weight, shots, thresholds, degrees_sum, J_eff, n):
-    weights = (1.0 / (1.0 + (2e-52) - J_eff)).astype(np.float64)
-
+def sample_for_solution(G_func, G_func_args_tuple, nodes, max_weight, shots, thresholds, degrees_sum, weights, n, dtype):
     solutions = np.empty((shots, n), dtype=np.bool_)
-    energies = np.empty(shots, dtype=np.float32)
+    energies = np.empty(shots, dtype=dtype)
 
     for s in prange(shots):
         # First dimension: Hamming weight
@@ -119,10 +117,10 @@ def sample_for_solution(G_func, G_func_args_tuple, nodes, max_weight, shots, thr
 
 
 @njit(parallel=True)
-def init_J_and_z(G_func, G_func_args_tuple, nodes):
+def init_J_and_z(G_func, G_func_args_tuple, nodes, dtype):
     n_qubits = len(nodes)
     degrees = np.empty(n_qubits, dtype=np.uint32)
-    J_eff = np.empty(n_qubits, dtype=np.float32)
+    J_eff = np.empty(n_qubits, dtype=dtype)
     J_max = -float("inf")
     G_max = -float("inf")
     for n in prange(n_qubits):
@@ -145,13 +143,17 @@ def init_J_and_z(G_func, G_func_args_tuple, nodes):
 
 
 @njit
-def cpu_footer(shots, quality, n_qubits, G_func, G_func_args_tuple, nodes):
-    J_eff, degrees, G_max = init_J_and_z(G_func, G_func_args_tuple, nodes)
-    hamming_prob = init_thresholds(n_qubits)
+def cpu_footer(shots, quality, n_qubits, G_func, G_func_args_tuple, nodes, dtype):
+    J_eff, degrees, G_max = init_J_and_z(G_func, G_func_args_tuple, nodes, dtype)
+    hamming_prob = init_thresholds(n_qubits, dtype)
 
     maxcut_hamming_cdf(n_qubits, J_eff, degrees, quality, hamming_prob)
 
-    best_solution, best_value = sample_for_solution(G_func, G_func_args_tuple, nodes, G_max, shots, hamming_prob, degrees.sum(), J_eff, n_qubits)
+    max_weight = degrees.sum()
+    degrees = None
+    J_eff = 1.0 / (1.0 + (2 ** -52) - J_eff)
+
+    best_solution, best_value = sample_for_solution(G_func, G_func_args_tuple, nodes, G_max, shots, hamming_prob, max_weight, J_eff, n_qubits, dtype)
 
     bit_string, l, r = get_cut(best_solution, nodes)
 
@@ -159,10 +161,10 @@ def cpu_footer(shots, quality, n_qubits, G_func, G_func_args_tuple, nodes):
 
 
 @njit
-def gpu_footer(shots, n_qubits, G_func, G_func_args_tuple, nodes, G_max, J_eff, degrees, hamming_prob):
+def gpu_footer(shots, n_qubits, G_func, G_func_args_tuple, nodes, G_max, weights, degrees, hamming_prob, max_weight, dtype):
     fix_cdf(hamming_prob)
 
-    best_solution, best_value = sample_for_solution(G_func, G_func_args_tuple, nodes, G_max, shots, hamming_prob, degrees.sum(), J_eff, n_qubits)
+    best_solution, best_value = sample_for_solution(G_func, G_func_args_tuple, nodes, G_max, shots, hamming_prob, max_weight, weights, n_qubits, dtype)
 
     bit_string, l, r = get_cut(best_solution, nodes)
 
@@ -175,7 +177,9 @@ def maxcut_tfim_streaming(
     G_func_args_tuple=None,
     quality=None,
     shots=None,
+    is_base_maxcut_gpu=True
 ):
+    dtype = opencl_context.dtype
     n_qubits = len(nodes)
 
     if n_qubits < 3:
@@ -199,19 +203,19 @@ def maxcut_tfim_streaming(
         # Number of measurement shots
         shots = n_qubits << quality
 
-    n_steps = 2 << quality
+    n_steps = n_qubits << quality
     grid_size = n_steps * n_qubits
 
-    if not (IS_OPENCL_AVAILABLE and grid_size >= 128):
-        return cpu_footer(shots, quality, n_qubits, G_func, G_func_args_tuple, nodes)
+    if (not is_base_maxcut_gpu) or not (IS_OPENCL_AVAILABLE and grid_size >= 128):
+        return cpu_footer(shots, quality, n_qubits, G_func, G_func_args_tuple, nodes, dtype)
 
-    J_eff, degrees, G_max = init_J_and_z(G_func, G_func_args_tuple, nodes)
+    J_eff, degrees, G_max = init_J_and_z(G_func, G_func_args_tuple, nodes, dtype)
 
     delta_t = 1.0 / n_steps
     tot_t = 2.0 * n_steps * delta_t
     h_mult = 2.0 / tot_t
 
-    args = np.empty(3, dtype=np.float32)
+    args = np.empty(3, dtype=dtype)
     args[0] = delta_t
     args[1] = tot_t
     args[2] = h_mult
@@ -232,7 +236,7 @@ def maxcut_tfim_streaming(
         args_buf, np.int32(n_qubits), J_buf, deg_buf, theta_buf
     )
 
-    hamming_prob = init_thresholds(n_qubits)
+    hamming_prob = init_thresholds(n_qubits, dtype)
 
     # Warp size is 32:
     group_size = n_qubits - 1
@@ -263,4 +267,8 @@ def maxcut_tfim_streaming(
     deg_buf = None
     theta_buf = None
 
-    return gpu_footer(shots, n_qubits, G_func, G_func_args_tuple, nodes, G_max, J_eff, degrees, hamming_prob)
+    max_weight = degrees.sum()
+    degrees = None
+    J_eff = 1.0 / (1.0 + (2 ** -52) - J_eff)
+
+    return gpu_footer(shots, n_qubits, G_func, G_func_args_tuple, nodes, G_max, J_eff, degrees, hamming_prob, max_weight, dtype)
