@@ -21,8 +21,8 @@ def get_cut(solution, nodes):
 
 
 @njit(parallel=True)
-def init_theta(h_mult, n_qubits, J_eff, degrees):
-    theta = np.empty(n_qubits, dtype=np.float32)
+def init_theta(h_mult, n_qubits, J_eff, degrees, dtype):
+    theta = np.empty(n_qubits, dtype=dtype)
     h_mult = abs(h_mult)
     for q in prange(n_qubits):
         J = J_eff[q]
@@ -41,9 +41,9 @@ def init_theta(h_mult, n_qubits, J_eff, degrees):
 
 
 @njit
-def init_thresholds(n_qubits):
+def init_thresholds(n_qubits, dtype):
     n_bias = n_qubits - 1
-    thresholds = np.empty(n_bias, dtype=np.float32)
+    thresholds = np.empty(n_bias, dtype=dtype)
     tot_prob = 0
     p = 1.0
     if n_qubits & 1:
@@ -62,7 +62,7 @@ def init_thresholds(n_qubits):
 
 
 @njit(parallel=True)
-def maxcut_hamming_cdf(n_qubits, J_func, degrees, quality, hamming_prob):
+def maxcut_hamming_cdf(n_qubits, J_func, degrees, quality, hamming_prob, dtype):
     if n_qubits < 2:
         hamming_prob.fill(0.0)
         return
@@ -86,8 +86,8 @@ def maxcut_hamming_cdf(n_qubits, J_func, degrees, quality, hamming_prob):
         t = step * delta_t
         tm1 = (step - 1) * delta_t
         h_t = h_mult * (tot_t - t)
-        bias = probability_by_hamming_weight(J_eff, h_t, z, theta_eff, t, n_qubits)
-        last_bias = probability_by_hamming_weight(J_eff, h_t, z, theta_eff, tm1, n_qubits)
+        bias = probability_by_hamming_weight(J_eff, h_t, z, theta_eff, t, n_qubits, dtype)
+        last_bias = probability_by_hamming_weight(J_eff, h_t, z, theta_eff, tm1, n_qubits, dtype)
         for i in range(n_bias):
             hamming_prob[i] += bias[i] - last_bias[i]
 
@@ -112,9 +112,7 @@ def fix_cdf(hamming_prob):
 
 
 @njit
-def probability_by_hamming_weight(J, h, z, theta, t, n_qubits):
-    bias = np.empty(n_qubits - 1, dtype=np.float32)
-
+def probability_by_hamming_weight(J, h, z, theta, t, n_qubits, dtype):
     # critical angle
     theta_c = np.arcsin(max(-1.0, min(1.0, abs(h) / (z * J))))
 
@@ -125,8 +123,9 @@ def probability_by_hamming_weight(J, h, z, theta, t, n_qubits):
     )
 
     if (p * n_qubits) >= 1024:
-        return bias
+        return np.zeros(n_qubits - 1, dtype=dtype)
 
+    bias = np.empty(n_qubits - 1, dtype=dtype)
     tot_n = 1.0 + 1.0 / pow(2.0, p * n_qubits)
     factor = pow(2.0, -p)
     n = 1.0
@@ -134,6 +133,10 @@ def probability_by_hamming_weight(J, h, z, theta, t, n_qubits):
         n *= factor
         bias[q - 1] = n
         tot_n += n
+
+    if np.isnan(tot_n) or np.isinf(tot_n):
+        return np.zeros(n_qubits - 1, dtype=dtype)
+
     bias /= tot_n
 
     if J > 0.0:
@@ -143,9 +146,10 @@ def probability_by_hamming_weight(J, h, z, theta, t, n_qubits):
 
 
 class OpenCLContext:
-    def __init__(self, p, a, c, q, i, m, b, s, k, l):
+    def __init__(self, p, a, d, c, q, i, m, b, s, k, l):
         self.MAX_GPU_PROC_ELEM = p
         self.IS_OPENCL_AVAILABLE = a
+        self.dtype = d
         self.ctx = c
         self.queue = q
         self.init_theta_kernel = i
@@ -163,12 +167,44 @@ IS_OPENCL_AVAILABLE = True
 ctx = None
 queue = None
 compute_units = None
+dtype = np.float32
 init_theta_kernel = None
 maxcut_hamming_cdf_kernel = None
 bootstrap_kernel = None
 bootstrap_sparse_kernel = None
 sample_for_solution_best_bitset_kernel = None
 sample_for_solution_best_bitset_sparse_kernel = None
+
+dtype_bits = int(os.getenv('PYQRACKISING_FPPOW', '5'))
+kernel_src = ''
+if dtype_bits <= 4:
+    dtype = np.float16
+    kernel_src += "#pragma OPENCL EXTENSION cl_khr_fp16 : enable\n"
+    kernel_src += "#define FP16 1\n"
+    kernel_src += "#define real1 half\n"
+    kernel_src += "#define qint short\n"
+    kernel_src += "#define EPSILON ((half)0.00097656f)\n"
+    kernel_src += "#define ZERO_R1 ((half)0.0f)\n"
+    kernel_src += "#define ONE_R1 ((half)1.0f)\n"
+    kernel_src += "#define TWO_R1 ((half)2.0f)\n"
+elif dtype_bits >= 6:
+    dtype = np.float64
+    kernel_src += "#pragma OPENCL EXTENSION cl_khr_fp64 : enable\n"
+    kernel_src += "#define real1 double\n"
+    kernel_src += "#define qint long\n"
+    kernel_src += "#define EPSILON DBL_EPSILON\n"
+    kernel_src += "#define ZERO_R1 0.0\n"
+    kernel_src += "#define ONE_R1 1.0\n"
+    kernel_src += "#define TWO_R1 2.0\n"
+else:
+    dtype = np.float32
+    kernel_src += "#define real1 float\n"
+    kernel_src += "#define qint int\n"
+    kernel_src += "#define EPSILON FLT_EPSILON\n"
+    kernel_src += "#define ZERO_R1 0.0f\n"
+    kernel_src += "#define ONE_R1 1.0f\n"
+    kernel_src += "#define TWO_R1 2.0f\n"
+
 try:
     import pyopencl as cl
     
@@ -178,7 +214,8 @@ try:
     compute_units = int(os.getenv('PYQRACKISING_MAX_GPU_PROC_ELEM', str(ctx.devices[0].get_info(cl.device_info.MAX_COMPUTE_UNITS))))
 
     # Load and build OpenCL kernels
-    kernel_src = f"#define MAX_PROC_ELEM {compute_units}\n"
+    kernel_src += f"#define MAX_PROC_ELEM {compute_units}\n"
+    kernel_src += f"#define TOP_N {os.getenv('PYQRACKISING_GPU_TOP_N', '32')}\n"
     kernel_src += open(os.path.dirname(os.path.abspath(__file__)) + "/kernels.cl").read()
     program = cl.Program(ctx, kernel_src).build()
     init_theta_kernel = program.init_theta
@@ -191,4 +228,4 @@ except ImportError:
     IS_OPENCL_AVAILABLE = False
     print("PyOpenCL not installed. (If you have any OpenCL accelerator devices with available ICDs, you might want to optionally install pyopencl.)")
 
-opencl_context = OpenCLContext(compute_units, IS_OPENCL_AVAILABLE, ctx, queue, init_theta_kernel, maxcut_hamming_cdf_kernel, bootstrap_kernel, bootstrap_sparse_kernel, sample_for_solution_best_bitset_kernel, sample_for_solution_best_bitset_sparse_kernel)
+opencl_context = OpenCLContext(compute_units, IS_OPENCL_AVAILABLE, dtype, ctx, queue, init_theta_kernel, maxcut_hamming_cdf_kernel, bootstrap_kernel, bootstrap_sparse_kernel, sample_for_solution_best_bitset_kernel, sample_for_solution_best_bitset_sparse_kernel)
