@@ -52,9 +52,9 @@ def bootstrap_worker(theta, G_data, G_rows, G_cols, indices):
 
 
 @njit(parallel=True)
-def bootstrap(best_theta, G_data, G_rows, G_cols, indices_array, k, min_energy):
+def bootstrap(best_theta, G_data, G_rows, G_cols, indices_array, k, min_energy, dtype):
     n = len(indices_array) // k
-    energies = np.empty(n, dtype=np.float32)
+    energies = np.empty(n, dtype=dtype)
     for i in prange(n):
         j = i * k
         energies[i] = bootstrap_worker(best_theta, G_data, G_rows, G_cols, indices_array[j : j + k])
@@ -74,6 +74,7 @@ def run_bootstrap_opencl(best_theta, G_data_buf, G_rows_buf, G_cols_buf, indices
     ctx = opencl_context.ctx
     queue = opencl_context.queue
     bootstrap_kernel = opencl_context.bootstrap_sparse_kernel
+    dtype = opencl_context.dtype
 
     n = best_theta.shape[0]
     combo_count = len(indices_array_np) // k
@@ -90,7 +91,7 @@ def run_bootstrap_opencl(best_theta, G_data_buf, G_rows_buf, G_cols_buf, indices
     args_buf = cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=args_np)
 
     # Allocate min_energy and min_index result buffers per workgroup
-    min_energy_host = np.empty(combo_count, dtype=np.float32)
+    min_energy_host = np.empty(combo_count, dtype=dtype)
     min_index_host = np.empty(combo_count, dtype=np.int32)
 
     min_energy_buf = cl.Buffer(ctx, mf.WRITE_ONLY, min_energy_host.nbytes)
@@ -99,11 +100,12 @@ def run_bootstrap_opencl(best_theta, G_data_buf, G_rows_buf, G_cols_buf, indices
     # Local memory allocation (1 float per work item)
     local_size = min(64, n)
     global_size = ((combo_count + local_size - 1) // local_size) * local_size
-    local_energy_buf = cl.LocalMemory(np.dtype(np.float32).itemsize * local_size)
+    local_energy_buf = cl.LocalMemory(np.dtype(dtype).itemsize * local_size)
     local_index_buf = cl.LocalMemory(np.dtype(np.int32).itemsize * local_size)
 
     # Set kernel args
     bootstrap_kernel.set_args(
+        np.random.randint(1<<32, dtype='uint32'),
         G_data_buf,
         G_rows_buf,
         G_cols_buf,
@@ -124,12 +126,12 @@ def run_bootstrap_opencl(best_theta, G_data_buf, G_rows_buf, G_cols_buf, indices
     queue.finish()
 
     # Find global minimum
-    min_idx = np.argmin(min_energy_host)
-    energy = min_energy_host[min_idx]
+    energy = min_energy_host.min()
     if min_energy < energy:
         return min_energy
 
-    best_i = min_index_host[min_idx]
+    choices = np.where(np.isclose(min_energy_host, energy, atol=2e-23, rtol=0))[0]
+    best_i = np.random.choice(choices) if len(choices) else np.argmin(min_energy_host)
 
     flip_index_start = best_i * k
     indices_to_flip = indices_array_np[flip_index_start : flip_index_start + k]
@@ -140,8 +142,8 @@ def run_bootstrap_opencl(best_theta, G_data_buf, G_rows_buf, G_cols_buf, indices
     return energy
 
 
-def to_scipy_sparse_upper_triangular(G, nodes, n_nodes):
-    lil = lil_matrix((n_nodes, n_nodes), dtype=np.float64)
+def to_scipy_sparse_upper_triangular(G, nodes, n_nodes, dtype):
+    lil = lil_matrix((n_nodes, n_nodes), dtype=dtype)
     for u in range(n_nodes):
         u_node = nodes[u]
         for v in range(u + 1, n_nodes):
@@ -157,7 +159,9 @@ def spin_glass_solver_sparse(
     quality=None,
     shots=None,
     best_guess=None,
-    is_alt_gpu_sampling=False
+    is_alt_gpu_sampling=False,
+    is_base_maxcut_gpu=True,
+    is_combo_maxcut_gpu=True
 ):
     nodes = None
     n_qubits = 0
@@ -196,10 +200,10 @@ def spin_glass_solver_sparse(
     elif isinstance(best_guess, list):
         bitstring = "".join(["1" if b else "0" for b in best_guess])
     else:
-        bitstring, _, _ = maxcut_tfim_sparse(G_m, quality=quality, shots=shots, is_alt_gpu_sampling=is_alt_gpu_sampling, is_g_buf_reused=True)
+        bitstring, _, _ = maxcut_tfim_sparse(G_m, quality=quality, shots=shots, is_alt_gpu_sampling=is_alt_gpu_sampling, is_g_buf_reused=True, is_base_maxcut_gpu=is_base_maxcut_gpu)
     best_theta = np.array([b == "1" for b in list(bitstring)], dtype=np.bool_)
 
-    if IS_OPENCL_AVAILABLE:
+    if is_combo_maxcut_gpu and IS_OPENCL_AVAILABLE:
         if not (opencl_context.G_data_buf is None):
             G_data_buf = opencl_context.G_data_buf
             G_rows_buf = opencl_context.G_rows_buf
@@ -231,10 +235,10 @@ def spin_glass_solver_sparse(
             else:
                 combos = combos_list[k - 1]
 
-            if IS_OPENCL_AVAILABLE:
+            if is_combo_maxcut_gpu and IS_OPENCL_AVAILABLE:
                 energy = run_bootstrap_opencl(best_theta, G_data_buf, G_rows_buf, G_cols_buf, combos, k, min_energy)
             else:
-                energy = bootstrap(best_theta, G_m.data, G_m.indptr, G_m.indices, combos, k, min_energy)
+                energy = bootstrap(best_theta, G_m.data, G_m.indptr, G_m.indices, combos, k, min_energy, dtype)
 
             if energy < min_energy:
                 min_energy = energy
